/********************************************************************
 * level2pbx.c
 * Part of maKs
 * Copyright Chris Benndorf, 1998
 * Released under GNU Public Library License
 * Best readeable with ":set tabstop=4"
 *
 * $Id: level2pbx.c,v 1.5 2000/01/13 21:49:57 root Exp root $
 *
 ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#include <sys/termios.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include <glob.h>
#include <log.h>
#include <cl_utils.h>
#include <level2pbx.h>

#include <../config.h>

#ifndef memcpy
#define memcpy(d,s,l)	(bcopy(s,d,l))
#endif

#define	CHKSUMVAL 0x8000

struct	termios _tser;
struct	termios _tsero;
char		_premature[256];
char		_lckfile[256];
int			_fd_serial;
FILE *		_fp_serial;
int			_receiving	= 0;
int			_sending	= 0;
char *		_pmwptr = _premature;
char *		_pmrptr = _premature;
int			_pmlen = 0;
int			_psent = 0;
int			_prcvd = 0;
int			_serrs = 0;
int			_rerrs = 0;
int			_sreq  = 0;
int			_rreq  = 0;
int			_numpacks  = 0;
time_t		_lastpreq;
PackEntry*	packHead = NULL;

int pbx_pck_response();

/********************************************************************
 * Checks if serial port is locked. If not it locks it
 * Parameters: name of serial port
 * Returns:    0 for OK, -1 for error
 ********************************************************************/
int pbx_lock(char *ser_dev)
{
	static int i,n,lock,pid;
	char buf[256];
	struct stat stt;

	if (LOCKDIR[0] && stat(LOCKDIR, &stt) == 0)
		sprintf(_lckfile, "%s/LCK..%s", LOCKDIR, basename(ser_dev));
	else
		_lckfile[0] = 0;

	if (_lckfile[0] && (lock = open(_lckfile, O_RDONLY)) >= 0) {
		n = read(lock, buf, 127);
		if (n > 0) {
			pid = -1;
			if (n == 4)
				/* Kermit-style lockfile. */
				pid = *(int *)buf;
			else {
				/* Ascii lockfile. */
				buf[n] = 0;
				sscanf(buf, "%d", &pid);
			}
			if (pid > 0 && kill((pid_t)pid, 0) < 0 && errno == ESRCH) {
				logerror("Lockfile is stale. Overriding it..\n");
				sleep(1);
				unlink(_lckfile);
			} else
				n = 0;
		}
		if (n == 0) {
			logerror("Device %s is locked.\n", ser_dev);
			return(-1);
		}
	}

	if ( _lckfile[0]) {
		/* Create lockfile compatible with UUCP-1.2 */
		if ((lock = open(_lckfile, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) {
			logerror(_lckfile);
			logerror("Cannot create lockfile. Sorry.\n");
			_lckfile[0] = 0;
			return(-1);
		}
		(void) umask(022);
		sprintf(buf, "%05d maximaX\n", getpid());
		write(lock, buf, strlen(buf));
		close(lock);
	}
	return(0);
}

/********************************************************************
 * Open serial port and initialize it
 * Parameters: name of serial port
 * Returns:    fd of serial port or -1 for error
 ********************************************************************/
int pbx_open(char *ser_dev, int serspeed)
{
	int ser;

	if (pbx_lock(ser_dev)) return(-1);
	ser=open(ser_dev,O_RDWR|O_NDELAY|O_EXCL);	
	if (ser==-1) {
		logerror("Cannot open %s, errno %d\n",ser_dev,errno);
		if (_lckfile[0]) unlink(_lckfile);
		return(-1);
	}

	tcgetattr(ser,&_tsero);
	_tser=_tsero;
	_tser.c_iflag &= ~(IGNBRK | IGNCR | INLCR | ICRNL | 
#ifdef __linux__
		IUCLC |
#endif
		IXANY | IXON | IXOFF | INPCK | ISTRIP);
	_tser.c_iflag |= (BRKINT | IGNPAR | IGNBRK);
	_tser.c_oflag &= ~OPOST;
	_tser.c_lflag &= ~(ICANON | ISIG | ECHO | ECHONL | ECHOE | ECHOK);
	_tser.c_cflag &= ~(HUPCL | CRTSCTS | CSIZE ) ;
	_tser.c_cflag |= CREAD | CRTSCTS | CS8 | CSTOPB |CLOCAL;
	_tser.c_cc[VMIN] = 1;
	_tser.c_cc[VTIME] = 0;
	cfsetospeed(&_tser,serspeed);
	cfsetispeed(&_tser,serspeed);
	tcsetattr(ser, TCSANOW, &_tser);

	_fd_serial = ser;
	_fp_serial = fdopen(ser, "w+");
	return(ser);
}

/********************************************************************
 * close serial port and delete lockfile
 * Parameters: fd for pbx
 ********************************************************************/
void pbx_close()
{
	int	serori = 0;

	fcntl(_fd_serial,F_SETFL,serori);
	_tser = _tsero;
	tcflush(_fd_serial,TCIOFLUSH);
	tcsetattr(_fd_serial,TCSANOW,&_tser);
	
	close(_fd_serial);
	if (_lckfile[0]) unlink(_lckfile);
	return;
}

/********************************************************************
 * Get file descriptor for PBX
 * Parameters: none
 * Returns:    fd
 ********************************************************************/
int pbx_fd()
{
	return(_fd_serial);
}

/********************************************************************
 * pbx_read reads data from pbx
 * Parameters: fd for pbx, buffer , length of buffer
 * returns:    amount of written data or -1
 ********************************************************************/
int pbx_read(BYTE *buf, int maxlen, int wait)
{
	static fd_set readseta;
	struct timeval timeout;
	int ret=0;
	int nsel=0;

	errno = 0;

	timeout.tv_sec = wait / 1000; 
	timeout.tv_usec= wait - ((wait / 1000) * 1000);

	FD_ZERO(&readseta);
	FD_SET(_fd_serial, &readseta);
	if ((nsel=select(_fd_serial + 1, &readseta, NULL, NULL, (wait) ? &timeout : NULL))) {
		if (FD_ISSET(_fd_serial, &readseta)) {
			ret=read(_fd_serial, buf, maxlen);
			logbuf(LOG_IOBYTES, "pbx_read", buf, ret);
			if (!ret) errno = ENOTCONN;
       			return (ret);
		}
	}
	return(ret);
}

/********************************************************************
 * writePBX writes data to pbx
 * Parameters: fd for pbx, buffer , length of buffer
 * returns:    amount of written data or -1
 ********************************************************************/
int pbx_write(BYTE *buf, int len)
{
	static fd_set writeseta;
	struct timeval timeout;
	int ret=0;

	errno = 0;

	timeout.tv_sec = 2;
	timeout.tv_usec = 0;

	logbuf(LOG_IOBYTES, "pbx_write", buf, len);
	
	FD_ZERO(&writeseta);
	FD_SET(_fd_serial, &writeseta);

	select(_fd_serial + 1, NULL, &writeseta, NULL, &timeout);
	if (FD_ISSET(_fd_serial, &writeseta)) {
		ret=write(_fd_serial, buf, len);
		return(ret);
	}
	return(-1);
}

void pbx_flush_in()
{
	tcflush(_fd_serial, TCIFLUSH);
}

/********************************************************************
 * Checksum algorythm for E-Packets
 * Parameters: buffer , length of buffer
 * returns:    16-bit checksum 
 ********************************************************************/
WORD calc_chksum(BYTE *buf, int len)
{
	register WORD	checks	= 0;
	register BYTE		*bp		= buf;
	register int		llen	= len - 3;

	for ( ; llen ; llen-- ) {

		if ( checks & 01 )
			checks = (WORD)(( checks >> 1 ) - CHKSUMVAL );
		else
			checks >>= 1;

		checks = (WORD)( checks + (WORD)*bp );
		bp++;
	}

	return(checks);
}

/********************************************************************
 * Calculates and fills in checksum for packet.
 *
 * Parameters: packet buffer , length of buffer
 ********************************************************************/
void fill_in_chksum(BYTE *buf, int len)
{
	WORD	cs;

	cs = calc_chksum(buf, len);

	buf[len - 3] = cs >> 8;
	buf[len - 2] = cs & BYTE_MASK;
}

/********************************************************************
 * Calculates and compares checksum for packet.
 *
 * Parameters: packet buffer , length of buffer
 * Returns:    >0 for equal, 0 for not equal
 ********************************************************************/
int check_chksum(BYTE *buf, int len)
{
	WORD	ccs;
	WORD	pcs;

	if (len < 5) return(FALSE);

	ccs = calc_chksum(buf, len);

	pcs = (buf[len - 3] << 8) | (buf[len - 2]);

	return((ccs == pcs));
}

/********************************************************************
 * While transmitting data from or to the pbx all bytes in the range
 * 0xE0 and 0xEF are transferred as two bytes (0xE0 0x0[0-F]) to
 * avoid conflicts.
 ********************************************************************/

/********************************************************************
 * de_E0_buffer translates all accurances of 0xE00[0-F] to 0xE[0-F]
 * Parameters: buffer to received data, length of data
 * Returns:    new length of data
 ********************************************************************/
int de_E0_buffer(char *buf, int len)
{
	register BYTE *bp	= buf;
	register BYTE *np	= bp;
	register int	ln	= len;
	register int	rt	= 0;

	for ( ; ln ; ln-- ) {
		if (*bp == CTI_ESC) {
			*np = (bp[0] | bp[1]);
			bp++;
			ln--;
		} else {
			*np = *bp;
		}
		bp++;
		np++;
		rt++;
	}
	return(rt);
}

/********************************************************************
 * en_E0_buffer translates all accurances of 0xE[0-F] to 0xE00[0-F]
 * Parameters: buffer to data, length of data, pointer to int of new len
 * Returns:    new data pointer, which must be freed by user
 ********************************************************************/
char *en_E0_buffer(char *buf, int len, int *nlen)
{
	char *dp	= buf;
	char *bp	= dp;
	char *rp;
	char *np;
	int E0_cnt	= 0;
	int llen	= len - 1; /*leave E8 alone */

	for ( ; llen ; llen-- ) {
		if ((*bp & HNIBBLE_MASK) == CTI_ESC)
			E0_cnt++;
		*bp++;
	}

	errno = 0;
	if ((rp = (char *) calloc(len + E0_cnt, 1)) == NULL) {
		logerror("en_E0_buffer: Error allocating %d bytes",len + E0_cnt);
		exit(1);
	}

	np = rp;
	bp = dp;
	llen = len - 1; /*leave E8 alone */

	for ( ; llen ; llen-- ) {
		if ((*bp & HNIBBLE_MASK ) == CTI_ESC){
			*np = CTI_ESC;
			np++;
			*np = (*bp & LNIBBLE_MASK);
		} else {
			*np = *bp;
		}
		np++;
		bp++;
	}

	*np = CTI_END_DATA;
	np++;

	*nlen = np - rp;

	return(rp);
}

/********************************************************************
 * Make up a proper E-Packet
 * Parameters: packet buffer, length of buffer
 * Returns:    buffer to complete packet
 ********************************************************************/
BYTE *pbx_pack_packet(BYTE *buf, int *len)
{
	BYTE	tmpbuf[512];
	BYTE	*tp		= tmpbuf;
	BYTE	*bp		= buf;
	BYTE	*np;
	int		llen	= *len;
	
	*tp++ = CTI_START_DATA;
	*tp++ = CTI_ESC;
	
	for ( ; llen ; llen-- ) 
		*tp++ = *bp++;

	*tp++ = 0;
	*tp++ = 0;
	*tp++ = CTI_END_DATA;

	llen = tp - tmpbuf;
	*len = llen;

	fill_in_chksum(tmpbuf, llen);

	np = en_E0_buffer(tmpbuf, llen, len);

	return(np);
}

/********************************************************************
 * Send out pbx E-command
 * Parameters: E-command to send
 ********************************************************************/
void pbx_send_E_cmd(BYTE c)
{
	pbx_write(&c, sizeof(char));
}

/********************************************************************
 * Tell PBX packet was ok (Checksum OK)
 * Parameters: none
 ********************************************************************/
void pbx_ack_packet()
{
	logprintf(LOG_HANDSHAKE,"R: CTI_BC_ACK\n");
	pbx_send_E_cmd(CTI_BC_ACK);
}

/********************************************************************
 * Tell PBX packet was broken (Checksum Wrong)
 * Parameters: none
 ********************************************************************/
void pbx_nack_packet()
{
	logprintf(LOG_HANDSHAKE,"R: CTI_BC_ERR\n");
	_rerrs++;
	pbx_send_E_cmd(CTI_BC_ERR);
}

/********************************************************************
 * Tell PBX to send data
 * Parameters: none
 ********************************************************************/
void pbx_cts_packet()
{
	logprintf(LOG_HANDSHAKE,"R: CTI_REC_READY\n");
	pbx_send_E_cmd(CTI_REC_READY);
	_rreq = 0;
}

/********************************************************************
 * Tell PBX reset_req was done
 * Parameters: none
 ********************************************************************/
void pbx_ack_reset()
{
	logprintf(LOG_HANDSHAKE,"R: CTI_RST_ACK\n");
	pbx_send_E_cmd(CTI_RST_ACK);
}

/********************************************************************
 * Ask PBX for reset
 * Parameters: none
 * Returns:    answer from pbx
 ********************************************************************/
int pbx_request_reset()
{
	int ret = 0;

	logprintf(LOG_HANDSHAKE,"R: CTI_RST_REQ\n");
	pbx_send_E_cmd(CTI_RST_REQ);

	return(ret);
}


/*****************************************************************
 * pbx_remove_pack
 *
 * Parameters: none
 * Returns:    void
 *****************************************************************/
void pbx_remove_pack()
{
	PackEntry	*lpe = packHead;

	if (lpe) {
		packHead = lpe->next;
		free(lpe->pack);
		free(lpe);
		_numpacks--;
	}
}

/*****************************************************************
 * pbx_try_read reads data, _not_ answering irretating CTI_SEND_REQs
 *
 * Parameters: same as pbx_read
 * Returns:    int (num bytes read)
 *****************************************************************/
int pbx_try_read(BYTE *buf, int maxlen, int wait) {
	int ret;
	BYTE rcv;

	if ((ret = pbx_read(buf, maxlen, wait))==0) {
		return(ret);
	}
	if (buf[0] == CTI_SEND_REQ) {
		logprintf(LOG_MSG,"Answering CTI_SEND_REQ delayed\n",rcv);
		_sreq++;
		return( pbx_read(buf, maxlen, wait) );
	}
	return(ret);
}

/*****************************************************************
 * pbx_send_packet
 *
 * Parameters: none
 * Returns:    void
 *****************************************************************/
int pbx_send_packet() {
	PackEntry	*lpe = packHead;
	cl_mgmt		*cl;
	int ret;
	BYTE rcv;

	// check PBX
	pbx_request_reset();
	if ((ret = pbx_try_read(&rcv, 1, 10000))==0) {
		logprintf(LOG_MSG,"Anlage antwortet nicht\n");
		return(-1);
	}
	if (rcv != CTI_RST_ACK) {
		logprintf(LOG_MSG,"Kein CTI_RST_ACK! (%02X)\n",rcv);
		return(-1);
	} else {
		logprintf(LOG_HANDSHAKE,"Got CTI_RST_ACK!\n",rcv);
		pbx_send_E_cmd(CTI_SEND_REQ);
		logprintf(LOG_HANDSHAKE,"R: CTI_SEND_REQ\n");
		if ((ret = pbx_try_read(&rcv, 1, 1000))==0) {
			logprintf(LOG_MSG,"Keine Antwort auf CTI_SEND_REQ!\n");
			return(-1);
		}
		if (rcv != CTI_REC_READY) {
			logprintf(LOG_MSG,"Kein CTI_REC_READY! (%02X)\n",rcv);
			return(-1);
		} else {
			logprintf(LOG_HANDSHAKE,"R: Sending packet\n");
			logbuf(LOG_SERBYTES, "pbx_send", lpe->pack, lpe->len);
			pbx_write(lpe->pack, lpe->len);
			_psent++;
			cl = Cl_find_serial();
			cl->last_pak = time(NULL);
			logprintf(LOG_HANDSHAKE," done.\n");
			if ((ret = pbx_try_read(&rcv, 1, 1000))==0) {
				logprintf(LOG_MSG,"Keine Antwort auf Paket!\n");
				return(-1);
			}
			switch (rcv) {
				case CTI_BC_ERR:
					_serrs++;
					logprintf(LOG_MSG,"CTI_BC_ERR:, sending packet again!\n");
					pbx_write(lpe->pack, lpe->len);
					if ((ret = pbx_try_read(&rcv, 1, 1000))==0) {
						logprintf(LOG_MSG,"Keine Antwort auf Paket!\n");
						return(-1);
					}
					switch (rcv) {
						case CTI_BC_ERR:
							_serrs++;
							logprintf(LOG_HANDSHAKE,"CTI_BC_ERR:, deleting packet! (after 2nd try)\n");
							return(-1);
							break;
						case CTI_BC_ACK:
							logprintf(LOG_HANDSHAKE,"CTI_BC_ACK: deleting packet (after 2nd try)!\n");
							return(0);
							break;
						default:
							logprintf(LOG_MSG,"Irretating Packet answer: %02X\n",rcv);
							return(-1);
							break;
					}
					pbx_remove_pack();
					break;
				case CTI_BC_ACK:
					logprintf(LOG_HANDSHAKE,"CTI_BC_ACK: deleting packet!\n");
					pbx_remove_pack();
					return(0);
					break;
				default:
					logprintf(LOG_MSG,"Irritating Packet answer: %02X\n",rcv);
					return(-1);
					break;
			}
		}
	}
	
}

/*****************************************************************
 * pbx_send_next
 *
 * Parameters: none
 * Returns:    void
 *****************************************************************/
int pbx_send_next() {
	PackEntry	*lpe = packHead;
	int ret=EOF;

	if (!lpe) return(ret);;

	ret = pbx_send_packet();
	if (_sreq) {
		_sreq = 0;
		pbx_cts_packet();
	}
	return(ret);
}

/*****************************************************************
 * pbx_send
 *
 * Parameters:
 * Returns:
 *****************************************************************/
int pbx_send(cl_mgmt *cl, BYTE *pak,int len)
{
	PackEntry	*npe;
	PackEntry	*ope;
	BYTE		*data;
	int			slen = len;

	errno = 0;
	if ((npe = (PackEntry *) calloc(sizeof(PackEntry),1)) == NULL) {
		logerror("Allocating new PackEntry!");
		exit(1);
	}

	errno = 0;
	if ((data = pbx_pack_packet(pak, &slen)) == NULL) {
		logerror("Allocating new Packet!");
		exit(1);
	}

	npe->pack = data;
	npe->len  = slen;
	npe->stat = 0;
	npe->next = NULL;

	_numpacks++;

	if (!packHead) {
		packHead = npe;
	} else {
		ope = packHead;
		while (ope->next)
			ope = ope->next;
		ope->next = npe;
	}
	
	return(CTI_BC_ACK);
}

/*****************************************************************
 * pbx_recpak receives packet from pbx. 
 *
 * Parameters: fd of pbx port, pointer to length
 * Returns:    pointer to received data
 *****************************************************************/
BYTE *pbx_recpak(cl_mgmt *cl, int *len)
{
	static BYTE	tmpbuf[512];
	int		llen;
	int		is_pack = 0;
	int		hskinpack = 0;
	int 	i;
	BYTE	*tbp = tmpbuf;
	BYTE	c = 0xAA;

	logprintf(LOG_HANDSHAKE,"A: Receiving packet...\n");

	while (c != CTI_END_DATA) {
		if ((c >= CTI_NUL ) && (c < CTI_END_DATA)){
			logprintf(LOG_MSG,"Handshake in the middle of packet (0x%02X)!!!\n",c);
			if (c == CTI_SEND_REQ)
				_sreq++;
		} else {
			*tbp++ = c;
		}
		if ((llen = pbx_read(&c, 1, 2000))!= 1) {
			errno = EIO;
			*len = 0;
			return(NULL);
		};
	}
	*tbp++ = c;

	logprintf(LOG_HANDSHAKE,"...done.\n");

	llen = tbp - tmpbuf;
	logbuf(LOG_SERBYTES, "pbx_receive", tmpbuf, llen);

	llen = de_E0_buffer(tmpbuf, llen); 

	if (check_chksum(tmpbuf, llen)) {

		/* Don't copy AA, E0, Checksum and E8 */
		llen -= 5;

		if ((tbp = (BYTE *) malloc(llen)) == NULL) {
			*len = 0;
			return(NULL);
		}

		memcpy(tbp, &tmpbuf[2], llen);

		if (tbp[6] + 7 > llen) {
			logprintf(LOG_MSG,"WRONG PLEN: PLEN=%d , MSGLEN=%d , CHKLEN=%d\n",
					llen, tbp[6],(tmpbuf[6] + 7));
			tbp[6]=0;
			logprintf(LOG_MSG,"This happens when there are too many CTI clients are connected!\n");
			logprintf(LOG_MSG,"This is a bug in the PABX firmware and I'm still trying to find a workaround.\n");

			pbx_ack_packet();
			errno = EIO;
			*len = 0;
			return(NULL);
		} else {
			pbx_ack_packet();
			_prcvd++;
		}

	} else {

		pbx_nack_packet();
		errno = EIO;
		*len = 0;
		return(NULL);

	}
	*len = llen;
	return(tbp);
}

/*****************************************************************
 * pbx_receive_packet goes into receive mode after CTI_SEND_REQ
 * and receives the whole packet.
 *
 * Parameters: fd of pbx port, pointer to length
 * Returns:    pointer to received data
 *****************************************************************/
BYTE *pbx_receive_packet(cl_mgmt *cl, int *len){
	int ret;
	BYTE rcv;

	// Send ClearToSend
	pbx_cts_packet();
	if ((ret = pbx_read(&rcv, 1, 1250))==0) {
		logprintf(LOG_MSG,"Packet-Receiver timed out!\n");
		return(NULL);
	}
	switch (rcv) {
		case CTI_START_DATA:
			return(pbx_recpak(cl, len));
		case CTI_SEND_REQ:
			return(pbx_receive_packet(cl, len));
		default:
			logprintf(LOG_MSG,"No CTI_START_DATA! (%02X)\n",rcv);
			return(NULL);
	}
	
	
}
/*****************************************************************
 * pbx_receive receives data from pbx. In case of a  CTI_SEND_REQ
 * byte, it receives the whole packet.
 *
 * Parameters: fd of pbx port, pointer to length
 * Returns:    pointer to received data
 *****************************************************************/
BYTE *pbx_receive(cl_mgmt *cl, int *len)
{
	int		llen;
	int		ret = EOF;
	int		is_pack = 0;
	BYTE	*tbp = NULL;
	BYTE	c;

	*len	= 0;
	errno	= 0;
	
	if ((llen = pbx_read(&c, 1, 500)) != 1) {
		return(NULL);
	}

	logbuf(LOG_SERBYTES, "pbx_receive", &c, llen);

	switch (c) {
		case CTI_START_DATA :
			return(pbx_recpak(cl, len));

		case CTI_SEND_REQ :
			logprintf(LOG_HANDSHAKE,"A: CTI_SEND_REQ\n");
			cl->last_pak = time(NULL);
			return(pbx_receive_packet(cl, len));
			break;
			
		case CTI_RST_REQ :
			logprintf(LOG_HANDSHAKE,"A: CTI_RST_REQ\n");
			cl->last_pak = time(NULL);
			pbx_ack_reset();
			break;

		default :
			logprintf(LOG_MSG,"A: UNEXPECTED BYTE %02X\n",c);
			break;
	}

	return(NULL);
}

/*****************************************************************
 * pbx_idle_proc should be called all 5 seconds if nothing else
 * is to done
 *
 * Parameters: none
 * Returns:    nothing
 *****************************************************************/
void pbx_idle_proc(cl_mgmt *this)
{
	cl_mgmt *cl;
	time_t	now = time(NULL);

	cl = Cl_find_serial();

	if (packHead) {
		pbx_send_next();
		return;
	}

	cl->pkg_sent = _psent;
	cl->pkg_rcvd = _prcvd;
	cl->pkg_serr = _serrs;
	cl->pkg_rerr = _rerrs;
	cl->pkg_left = _numpacks;
 
}

/*****************************************************************
 * pbx_init_proc should be called once after program is startet up
 *
 * Parameters: none
 * Returns:    nothing
 *****************************************************************/
cl_mgmt *pbx_init_proc(char *ser_dev, int serspeed)
{
	cl_mgmt *this;

	if (pbx_open(ser_dev, serspeed) == -1) exit(1);

	if ((this = Cl_create())==NULL) exit(1);
	Cl_init(this, _fd_serial, CTI_CL_SERIAL, "serial", "Timm");
	this->idleProc	= &pbx_idle_proc;
	this->readProc	= &pbx_receive;
	this->writeProc	= &pbx_send;

	return(this);
}

